from django.contrib.auth.decorators import login_required
from django.views.generic import TemplateView

from .backup_codes import BackupCodesView
from .login import KagiLoginView, VerifySecondFactorView
from .totp_devices import AddTOTPDeviceView, TOTPDeviceManagementView
from .webauthn_keys import AddWebAuthnKeyView, KeyManagementView


class TwoFactorSettingsView(TemplateView):
    template_name = "kagi/two_factor_settings.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["webauthn_enabled"] = self.request.user.webauthn_keys.exists()
        context["backup_codes_count"] = self.request.user.backup_codes.count()
        context["totp_enabled"] = self.request.user.totp_devices.exists()
        return context


add_webauthn_key = AddWebAuthnKeyView.as_view()
verify_second_factor = VerifySecondFactorView.as_view()
login = KagiLoginView.as_view()
keys = login_required(KeyManagementView.as_view())
two_factor_settings = login_required(TwoFactorSettingsView.as_view())
backup_codes = login_required(BackupCodesView.as_view())
add_totp = login_required(AddTOTPDeviceView.as_view())
totp_devices = login_required(TOTPDeviceManagementView.as_view())



# Generated by Django 2.2.4 on 2019-08-08 12:43

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="TOTPDevice",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("last_used_at", models.DateTimeField(null=True)),
                ("key", models.BinaryField()),
                ("last_t", models.PositiveIntegerField(null=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="totp_devices",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="BackupCode",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=8)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="backup_codes",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"unique_together": {("user", "code")}},
        ),
        migrations.CreateModel(
            name="WebAuthnKey",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("last_used_at", models.DateTimeField(blank=True, null=True)),
                ("key_name", models.CharField(max_length=64)),
                ("public_key", models.TextField(unique=True)),
                ("ukey", models.TextField(unique=True)),
                ("credential_id", models.TextField(unique=True)),
                ("sign_count", models.IntegerField()),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="webauthn_keys",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.contrib import admin
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _

from .models import WebAuthnKey


def make_login_view(view_class):
    def login(self, request, extra_context=None):
        """
        Displays the login form for the given HttpRequest.
        """
        if request.method == "GET" and self.has_permission(request):
            # Already logged-in, redirect to admin index
            index_path = reverse("admin:index", current_app=self.name)
            return HttpResponseRedirect(index_path)

        # Since this module gets imported in the application's root package,
        # it cannot import models from other applications at the module level,
        # and django.contrib.admin.forms eventually imports User.
        from django.contrib.admin.forms import AdminAuthenticationForm

        context = dict(
            self.each_context(request),
            title=_("Log in"),
            app_path=request.get_full_path(),
            username=request.user.get_username(),
        )
        if (
            REDIRECT_FIELD_NAME not in request.GET
            and REDIRECT_FIELD_NAME not in request.POST
        ):
            context[REDIRECT_FIELD_NAME] = reverse("admin:index", current_app=self.name)
        context.update(extra_context or {})

        defaults = {
            "extra_context": context,
            "authentication_form": self.login_form or AdminAuthenticationForm,
            "template_name": self.login_template or "admin/login.html",
        }
        request.current_app = self.name
        return view_class.as_view(**defaults)(request)

    return login


def monkeypatch_admin(view_class=None):
    if view_class is None:
        from kagi.views.login import KagiLoginView

        view_class = KagiLoginView

    from django.contrib.admin.sites import AdminSite

    AdminSite.login = make_login_view(view_class)


@admin.register(WebAuthnKey)
class WebAuthnKeyAdmin(admin.ModelAdmin):
    pass



from unittest import mock

from django.contrib.auth.models import User
from django.urls import reverse

import pytest

from .. import settings
from ..forms import KeyRegistrationForm
from ..models import WebAuthnKey


def test_list_webauthn_keys(admin_client):
    response = admin_client.get(reverse("kagi:webauthn-keys"))
    assert list(response.context_data["webauthnkey_list"]) == []
    assert response.status_code == 200


def test_webauthn_keys_str_return_the_username(admin_client):
    user = User.objects.get(pk=1)
    key = user.webauthn_keys.create(key_name="SoloKey", sign_count=0)
    assert str(key) == "admin - SoloKey"


def test_add_webauthn_key(admin_client):
    response = admin_client.get(reverse("kagi:add-webauthn-key"))
    assert response.status_code == 200
    assert isinstance(response.context_data["form"], KeyRegistrationForm)


def test_totp_device_deletion_works(admin_client):
    user = User.objects.get(pk=1)
    key = user.webauthn_keys.create(key_name="SoloKey", sign_count=0)

    response = admin_client.get(reverse("kagi:webauthn-keys"))
    assert response.status_code == 200
    assert len(response.context_data["webauthnkey_list"]) == 1
    response = admin_client.post(
        reverse("kagi:webauthn-keys"), {"delete": "checked", "key_id": key.pk}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:webauthn-keys")
    assert WebAuthnKey.objects.count() == 0


# Testing view begin activate
def test_begin_activate_return_user_credential_options(admin_client):
    ukey = "Q3sM6zbLYAssRO7g5BM7"
    with mock.patch("kagi.views.api.util.generate_ukey", return_value=ukey):
        response = admin_client.post(
            reverse("kagi:begin-activate"), {"key_name": "SoloKey"}
        )
    assert response.status_code == 200
    credential_options = response.json()
    assert "challenge" in credential_options
    assert credential_options["rp"] == {
        "name": settings.RELYING_PARTY_NAME,
        "id": settings.RELYING_PARTY_ID,
    }
    assert credential_options["user"] == {
        "id": ukey,
        "name": "admin",
        "displayName": "",
        "icon": settings.WEBAUTHN_ICON_URL,
    }

    assert "pubKeyCredParams" in credential_options
    assert credential_options["extensions"] == {"webauthn.loc": True}


def test_begin_activate_fails_if_key_name_is_missing(admin_client):
    response = admin_client.post(reverse("kagi:begin-activate"), {"key_name": ""})
    assert response.status_code == 400
    assert response.json() == {"errors": {"key_name": ["This field is required."]}}


# Testing view verify credential info
def test_webauthn_verify_credential_info(admin_client):
    # Setup the session
    response = admin_client.post(
        reverse("kagi:begin-activate"), {"key_name": "SoloKey"}
    )
    credential_options = response.json()
    challenge = credential_options["challenge"]

    trusted_attestation_cert_required = (
        settings.WEBAUTHN_TRUSTED_ATTESTATION_CERT_REQUIRED
    )
    self_attestation_permitted = settings.WEBAUTHN_SELF_ATTESTATION_PERMITTED
    none_attestation_permitted = settings.WEBAUTHN_NONE_ATTESTATION_PERMITTED

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_registration_response = (
            mocked_webauthn.WebAuthnRegistrationResponse.return_value
        )
        verify = webauthn_registration_response.verify.return_value
        verify.public_key.decode.return_value = "public-key"
        verify.credential_id.decode.return_value = "credential-id"
        verify.sign_count = 0

        response = admin_client.post(
            reverse("kagi:verify-credential-info"), {"registration": "payload"}
        )
    mocked_webauthn.WebAuthnRegistrationResponse.assert_called_with(
        settings.RELYING_PARTY_ID,
        "http://testserver",
        {"registration": ["payload"]},
        challenge,
        settings.WEBAUTHN_TRUSTED_CERTIFICATES,
        trusted_attestation_cert_required,
        self_attestation_permitted,
        none_attestation_permitted,
        uv_required=False,  # User validation
    )

    webauthn_registration_response.verify.assert_called_once()

    assert response.status_code == 200
    assert response.json() == {"success": "User successfully registered."}


def test_webauthn_verify_credential_info_fails_if_registration_is_invalid(admin_client):
    # Setup the session
    response = admin_client.post(
        reverse("kagi:begin-activate"), {"key_name": "SoloKey"}
    )

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_registration_response = (
            mocked_webauthn.WebAuthnRegistrationResponse.return_value
        )
        verify = webauthn_registration_response.verify
        verify.side_effect = ValueError("An error occurred")

        response = admin_client.post(
            reverse("kagi:verify-credential-info"), {"registration": "payload"}
        )

    assert response.status_code == 400
    assert response.json() == {"fail": "Registration failed. Error: An error occurred"}


def test_webauthn_verify_credential_info_fails_if_credential_id_already_exists(
    admin_client,
):
    # Setup the session
    response = admin_client.post(
        reverse("kagi:begin-activate"), {"key_name": "SoloKey"}
    )

    # Create the WebAuthnKey
    user = User.objects.get(pk=1)
    user.webauthn_keys.create(
        key_name="SoloKey", sign_count=0, credential_id="credential-id"
    )

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_registration_response = (
            mocked_webauthn.WebAuthnRegistrationResponse.return_value
        )
        verify = webauthn_registration_response.verify.return_value
        verify.credential_id.decode.return_value = "credential-id"

        response = admin_client.post(
            reverse("kagi:verify-credential-info"), {"registration": "payload"}
        )

    assert response.status_code == 400
    assert response.json() == {"fail": "Credential ID already exists."}


# Testing view begin assertion
@pytest.mark.django_db
def test_begin_assertion_return_user_credential_options(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.webauthn_keys.create(
        key_name="SoloKey 1",
        sign_count=0,
        credential_id="credential-id-1",
        ukey="abcd",
        public_key="pubkey1",
    )
    user.webauthn_keys.create(
        key_name="SoloKey 2",
        sign_count=0,
        credential_id="credential-id-2",
        ukey="efgh",
        public_key="pubkey2",
    )

    ukey = "Q3sM6zbLYAssRO7g5BM7"
    challenge = "k31d65xGDFb0VUq4MEMXmWpuWkzPs889"

    with mock.patch("kagi.views.api.util.generate_ukey", return_value=ukey):
        with mock.patch(
            "kagi.views.api.util.generate_challenge", return_value=challenge
        ):
            # We authenticate with username/password
            response = client.post(
                reverse("kagi:login"), {"username": "admin", "password": "admin"}
            )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        assertion_dict = {
            "challenge": "tOOk7MPjGWlezrP6o6tGOXSH0ZesUREO",
            "allowCredentials": [
                {
                    "type": "public-key",
                    "id": "ePqP9Mi...512GSYg",
                    "transports": ["usb", "nfc", "ble", "internal"],
                },
                {
                    "type": "public-key",
                    "id": "qhibXokRKbPA...O1WW7nF",
                    "transports": ["usb", "nfc", "ble", "internal"],
                },
            ],
            "rpId": "localhost",
            "timeout": 60000,
        }
        mocked_webauthn.WebAuthnAssertionOptions.return_value.assertion_dict = (
            assertion_dict
        )
        response = client.post(reverse("kagi:begin-assertion"))

    assert response.status_code == 200
    assert response.json() == assertion_dict


# Testing view verify assertion
@pytest.mark.django_db
def test_verify_assertion_validates_the_user_webauthn_key(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.webauthn_keys.create(
        key_name="SoloKey",
        sign_count=0,
        credential_id="credential-id",
        ukey="abcd",
        public_key="pubkey",
    )
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    # We authenticate with username/password
    challenge = "k31d65xGDFb0VUq4MEMXmWpuWkzPs889"

    with mock.patch("kagi.views.api.util.generate_challenge", return_value=challenge):
        response = client.post(reverse("kagi:begin-assertion"))

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_assertion_response = (
            mocked_webauthn.WebAuthnAssertionResponse.return_value
        )
        verify = webauthn_assertion_response.verify
        verify.return_value = 1

        response = client.post(
            reverse("kagi:verify-assertion"),
            {"id": "credential-id", "assertion": "payload"},
        )
    mocked_webauthn.WebAuthnUser.assert_called_with(
        "abcd",
        "admin",
        "",
        settings.WEBAUTHN_ICON_URL,
        "credential-id",
        "pubkey",
        0,
        settings.RELYING_PARTY_ID,
    )

    webauthn_user = mocked_webauthn.WebAuthnUser.return_value
    webauthn_assertion_response = mocked_webauthn.WebAuthnAssertionResponse
    webauthn_assertion_response.assert_called_with(
        webauthn_user,
        {"id": ["credential-id"], "assertion": ["payload"]},
        challenge,
        "http://testserver",
        uv_required=False,
    )

    assert response.status_code == 200
    assert response.json() == {
        "success": "Successfully authenticated as admin",
        "redirect_to": reverse("kagi:two-factor-settings"),
    }

    # Are we truly logged in?
    response = client.get(reverse("kagi:two-factor-settings"))
    assert response.status_code == 200


# Testing view verify assertion
@pytest.mark.django_db
def test_verify_assertion_fails_if_missing_user_webauthn_key(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.webauthn_keys.create(
        key_name="SoloKey",
        sign_count=0,
        credential_id="wrong-id",
        ukey="abcd",
        public_key="pubkey",
    )
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    # We authenticate with username/password
    challenge = "k31d65xGDFb0VUq4MEMXmWpuWkzPs889"

    with mock.patch("kagi.views.api.util.generate_challenge", return_value=challenge):
        response = client.post(reverse("kagi:begin-assertion"))

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_assertion_response = (
            mocked_webauthn.WebAuthnAssertionResponse.return_value
        )
        verify = webauthn_assertion_response.verify
        verify.return_value = 1

        response = client.post(
            reverse("kagi:verify-assertion"),
            {"id": "credential-id", "assertion": "payload"},
        )
    assert response.status_code == 400
    assert response.json() == {"fail": "Key does not exist."}


@pytest.mark.django_db
def test_verify_assertion_validates_the_assertion(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.webauthn_keys.create(
        key_name="SoloKey",
        sign_count=0,
        credential_id="credential-id",
        ukey="abcd",
        public_key="pubkey",
    )
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    # We authenticate with username/password
    challenge = "k31d65xGDFb0VUq4MEMXmWpuWkzPs889"

    response = client.get(reverse("kagi:verify-second-factor"))
    assert response.status_code == 200

    with mock.patch("kagi.views.api.util.generate_challenge", return_value=challenge):
        response = client.post(reverse("kagi:begin-assertion"))

    with mock.patch("kagi.views.api.webauthn") as mocked_webauthn:
        webauthn_assertion_response = (
            mocked_webauthn.WebAuthnAssertionResponse.return_value
        )
        verify = webauthn_assertion_response.verify
        verify.side_effect = ValueError("An error occurred")

        response = client.post(
            reverse("kagi:verify-assertion"),
            {"id": "credential-id", "assertion": "payload"},
        )

    assert response.status_code == 400
    assert response.json() == {"fail": "Assertion failed. Error: An error occurred"}



from django.contrib.auth.models import User
from django.urls import reverse


def test_mfa_settings_page_displays_mfa_statuses(admin_client):
    response = admin_client.get(reverse("kagi:two-factor-settings"))

    assert response.context_data["webauthn_enabled"] is False
    assert response.context_data["totp_enabled"] is False
    assert response.context_data["backup_codes_count"] == 0


def test_mfa_settings_page_knows_when_webauthn_is_enabled(admin_client):
    user = User.objects.get(pk=1)

    user.webauthn_keys.create(key_name="SoloKey", sign_count=0)
    response = admin_client.get(reverse("kagi:two-factor-settings"))

    assert response.context_data["webauthn_enabled"] is True
    assert response.context_data["totp_enabled"] is False
    assert response.context_data["backup_codes_count"] == 0


def test_mfa_settings_page_knows_when_totp_is_enabled(admin_client):
    user = User.objects.get(pk=1)

    user.totp_devices.create()
    response = admin_client.get(reverse("kagi:two-factor-settings"))

    assert response.context_data["webauthn_enabled"] is False
    assert response.context_data["totp_enabled"] is True
    assert response.context_data["backup_codes_count"] == 0


def test_mfa_settings_page_knows_how_to_count_backup_codes(admin_client):
    user = User.objects.get(pk=1)

    user.backup_codes.create_backup_code()
    response = admin_client.get(reverse("kagi:two-factor-settings"))

    assert response.context_data["webauthn_enabled"] is False
    assert response.context_data["totp_enabled"] is False
    assert response.context_data["backup_codes_count"] == 1



from base64 import b32decode, b32encode
from collections import OrderedDict
from io import BytesIO
import os
from urllib.parse import quote

from django.contrib import messages
from django.contrib.sites.shortcuts import get_current_site
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse, reverse_lazy
from django.utils.http import url_has_allowed_host_and_scheme, urlencode
from django.utils.translation import gettext as _
from django.views.generic import FormView, ListView

import qrcode
from qrcode.image.svg import SvgPathFillImage

from ..constants import SESSION_TOTP_SECRET_KEY
from ..forms import TOTPForm
from ..models import TOTPDevice
from .mixin import OriginMixin


class AddTOTPDeviceView(OriginMixin, FormView):
    form_class = TOTPForm
    template_name = "kagi/totp_device.html"
    success_url = reverse_lazy("kagi:totp-devices")

    def get(self, request, *args: str, **kwargs):
        # When opening the view with a GET request, we treat it as a "add new
        # device" request. There, we create a new TOTP secret and put it into
        # the current user's session. Upon POST, the secret is read from the
        # session again.
        # Once a new TOTP device was successfully added, we'll drop the secret
        # from the session.
        # This approach allows to re-enter the token if mistyped, while keeping
        # the same TOTP device setup on the TOTP generator.
        self.secret = self.gen_secret()
        request.session[SESSION_TOTP_SECRET_KEY] = self.secret
        return super().get(request, *args, **kwargs)

    def post(self, request, *args: str, **kwargs):
        # Try to get the TOTP secret from the session. If the secret doesn't
        # exist, redirect to the view again, to configure a new TOTP secret.
        self.secret = request.session.get(SESSION_TOTP_SECRET_KEY, None)
        if not self.secret:
            messages.error(request, _("Missing TOTP secret. Please try again."))
            return redirect(request.path)

        return super().post(request, *args, **kwargs)

    def gen_secret(self):
        return b32encode(os.urandom(20)).decode()

    def get_otpauth_url(self, secret):
        issuer = get_current_site(self.request).name

        params = OrderedDict([("secret", secret), ("digits", 6), ("issuer", issuer)])

        return "otpauth://totp/{issuer}:{username}?{params}".format(
            issuer=quote(issuer),
            username=quote(self.request.user.get_username()),
            params=urlencode(params),
        )

    def get_qrcode(self, data):
        img = qrcode.make(data, image_factory=SvgPathFillImage)
        buf = BytesIO()
        img.save(buf)
        return buf.getvalue().decode("utf-8")

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["base32_key"] = self.secret
        kwargs["otpauth"] = self.get_otpauth_url(self.secret)
        kwargs["qr_svg"] = self.get_qrcode(kwargs["otpauth"])
        return kwargs

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(
            user=self.request.user, request=self.request, appId=self.get_origin()
        )
        return kwargs

    def form_valid(self, form):
        device = TOTPDevice(user=self.request.user, key=b32decode(self.secret))
        if device.validate_token(form.cleaned_data["token"]):
            del self.request.session[SESSION_TOTP_SECRET_KEY]
            device.save()
            messages.success(self.request, _("Device added."))
            return super().form_valid(form)
        else:
            assert not device.pk
            form.add_error("token", TOTPForm.INVALID_ERROR_MESSAGE)
            return self.form_invalid(form)

    def form_invalid(self, form):
        # Should this go in Django's FormView?!
        # <https://code.djangoproject.com/ticket/25548>
        return self.render_to_response(self.get_context_data(form=form))

    def get_success_url(self):
        if "next" in self.request.GET and url_has_allowed_host_and_scheme(
            self.request.GET["next"], allowed_hosts=[self.request.get_host()]
        ):
            return self.request.GET["next"]
        else:
            return super().get_success_url()


class TOTPDeviceManagementView(ListView):
    template_name = "kagi/totpdevice_list.html"

    def get_queryset(self):
        return self.request.user.totp_devices.all()

    def post(self, request):
        assert "delete" in self.request.POST
        device = get_object_or_404(
            self.get_queryset(), pk=self.request.POST["device_id"]
        )
        device.delete()
        messages.success(request, _("Device removed."))
        return HttpResponseRedirect(reverse("kagi:totp-devices"))



from django import forms
from django.utils import timezone
from django.utils.translation import gettext_lazy as _


class SecondFactorForm(forms.Form):
    def __init__(self, *args, **kwargs):
        self.user = kwargs.pop("user")
        self.request = kwargs.pop("request")
        self.appId = kwargs.pop("appId")
        return super().__init__(*args, **kwargs)


class BackupCodeForm(SecondFactorForm):
    INVALID_ERROR_MESSAGE = _("That is not a valid backup code.")

    code = forms.CharField(
        label=_("Code"), widget=forms.TextInput(attrs={"autocomplete": "off"})
    )

    def validate_second_factor(self):
        count, _ = self.user.backup_codes.filter(
            code=self.cleaned_data["code"]
        ).delete()
        if count == 0:
            self.add_error("code", self.INVALID_ERROR_MESSAGE)
            return False
        elif count == 1:
            return True


class TOTPForm(SecondFactorForm):
    INVALID_ERROR_MESSAGE = _("That token is invalid.")

    token = forms.CharField(
        min_length=6,
        max_length=6,
        label=_("Token"),
        widget=forms.TextInput(attrs={"autocomplete": "off"}),
    )

    def validate_second_factor(self):
        for device in self.user.totp_devices.all():
            if device.validate_token(self.cleaned_data["token"]):
                device.last_used_at = timezone.now()
                device.save()
                return True
        self.add_error("token", self.INVALID_ERROR_MESSAGE)
        return False


class KeyRegistrationForm(forms.Form):
    key_name = forms.CharField(label=_("Key name"))



from django.conf import settings
from django.contrib import auth
from django.contrib.auth import load_backend
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.views import LoginView
from django.http import HttpResponseRedirect
from django.shortcuts import resolve_url
from django.urls import reverse
from django.utils.http import url_has_allowed_host_and_scheme, urlencode
from django.views.generic import TemplateView

from ..forms import BackupCodeForm, SecondFactorForm, TOTPForm
from .mixin import OriginMixin


class KagiLoginView(LoginView):
    form_class = AuthenticationForm
    template_name = "kagi/login.html"

    @property
    def is_admin(self):
        return self.template_name == "admin/login.html"

    def requires_two_factor(self, user):
        return user.webauthn_keys.exists() or user.totp_devices.exists()

    def form_valid(self, form):
        user = form.get_user()
        if not self.requires_two_factor(user):
            # no keys registered, use single-factor auth
            return super().form_valid(form)
        else:
            self.request.session["kagi_pre_verify_user_pk"] = user.pk
            self.request.session["kagi_pre_verify_user_backend"] = user.backend

            verify_url = reverse("kagi:verify-second-factor")
            redirect_to = self.request.POST.get(
                auth.REDIRECT_FIELD_NAME,
                self.request.GET.get(auth.REDIRECT_FIELD_NAME, ""),
            )
            params = {}
            if url_has_allowed_host_and_scheme(
                url=redirect_to,
                allowed_hosts=[self.request.get_host()],
                require_https=True,
            ):
                params[auth.REDIRECT_FIELD_NAME] = redirect_to
            if self.is_admin:
                params["admin"] = 1
            if params:
                verify_url += "?" + urlencode(params)

            return HttpResponseRedirect(verify_url)

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs[auth.REDIRECT_FIELD_NAME] = self.request.GET.get(
            auth.REDIRECT_FIELD_NAME, ""
        )
        kwargs.update(self.kwargs.get("extra_context", {}))
        return kwargs


class VerifySecondFactorView(OriginMixin, TemplateView):
    template_name = "kagi/verify_second_factor.html"

    @property
    def form_classes(self):
        ret = {}
        if self.user.webauthn_keys.exists():
            ret["webauthn"] = SecondFactorForm
        if self.user.backup_codes.exists():
            ret["backup"] = BackupCodeForm
        if self.user.totp_devices.exists():
            ret["totp"] = TOTPForm

        return ret

    def get_user(self):
        try:
            user_id = self.request.session["kagi_pre_verify_user_pk"]
            backend_path = self.request.session["kagi_pre_verify_user_backend"]
            assert backend_path in settings.AUTHENTICATION_BACKENDS
            backend = load_backend(backend_path)
            user = backend.get_user(user_id)
            if user is not None:
                user.backend = backend_path
            return user
        except (KeyError, AssertionError):
            return None

    def dispatch(self, request, *args, **kwargs):
        self.user = self.get_user()
        if self.user is None:
            return HttpResponseRedirect(reverse("kagi:login"))
        return super().dispatch(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        forms = self.get_forms()
        form = forms[request.POST["type"]]
        if form.is_valid() and form.validate_second_factor():
            return self.form_valid(form, forms)
        else:
            return self.form_invalid(forms)

    def form_invalid(self, forms):
        return self.render_to_response(self.get_context_data(forms=forms))

    def get_form_kwargs(self):
        return {"user": self.user, "request": self.request, "appId": self.get_origin()}

    def get_forms(self):
        kwargs = self.get_form_kwargs()
        if self.request.method == "GET":
            forms = {key: form(**kwargs) for key, form in self.form_classes.items()}
        else:
            method = self.request.POST["type"]
            forms = {
                key: form(**kwargs)
                for key, form in self.form_classes.items()
                if key != method
            }
            forms[method] = self.form_classes[method](self.request.POST, **kwargs)
        return forms

    def get_context_data(self, **kwargs):
        if "forms" not in kwargs:
            kwargs["forms"] = self.get_forms()
        kwargs = super().get_context_data(**kwargs)
        if self.request.GET.get("admin"):
            kwargs["base_template"] = "admin/base_site.html"
        else:
            kwargs["base_template"] = "base.html"
        kwargs["user"] = self.user
        return kwargs

    def form_valid(self, form, forms):
        del self.request.session["kagi_pre_verify_user_pk"]
        del self.request.session["kagi_pre_verify_user_backend"]

        auth.login(self.request, self.user)

        redirect_to = self.request.POST.get(
            auth.REDIRECT_FIELD_NAME, self.request.GET.get(auth.REDIRECT_FIELD_NAME, "")
        )
        if not url_has_allowed_host_and_scheme(
            url=redirect_to, allowed_hosts=[self.request.get_host()]
        ):
            redirect_to = resolve_url(settings.LOGIN_REDIRECT_URL)
        return HttpResponseRedirect(redirect_to)



from ..util import get_origin


def test_get_origin(rf):
    request = rf.get("/")
    origin = get_origin(request)
    assert origin == "http://testserver", "Origin should be 'testserver' over HTTP"



"""
This is inspired by addstatictoken in django-otp, which is licensed as follows:

    Copyright (c) 2012, Peter Sagerson
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    - Redistributions of source code must retain the above copyright notice, this
      list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    help = "Adds a single backup code to the given user."

    def add_arguments(self, parser):
        parser.add_argument("username")
        parser.add_argument(
            "--code",
            default=None,
            help="The code to add. If omitted, one will be randomly generated.",
        )

    def handle(self, *args, **options):
        user = get_user_model().objects.get_by_natural_key(options["username"])
        try:
            backupcode_obj = user.backup_codes.create_backup_code(options["code"])
        except Exception:
            raise CommandError("This code already exists.")
        print(backupcode_obj.code, file=self.stdout)



from django.urls import path

from . import views
from .views import api

app_name = "kagi"

urlpatterns = [
    path(
        "verify-second-factor/",
        views.verify_second_factor,
        name="verify-second-factor",
    ),
    path("login/", views.login, name="login"),
    path("keys/", views.keys, name="webauthn-keys"),
    path("add-webauthn-key/", views.add_webauthn_key, name="add-webauthn-key"),
    path("two-factor-settings/", views.two_factor_settings, name="two-factor-settings"),
    path("backup-codes/", views.backup_codes, name="backup-codes"),
    path("add-totp-device/", views.add_totp, name="add-totp"),
    path("totp-devices/", views.totp_devices, name="totp-devices"),
    path("api/begin-activate/", api.webauthn_begin_activate, name="begin-activate"),
    path(
        "api/verify-credential-info/",
        api.webauthn_verify_credential_info,
        name="verify-credential-info",
    ),
    path("api/begin-assertion/", api.webauthn_begin_assertion, name="begin-assertion"),
    path(
        "api/verify-assertion/",
        api.webauthn_verify_assertion,
        name="verify-assertion",
    ),
]



from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext as _
from django.views.generic import ListView, TemplateView

from ..forms import KeyRegistrationForm
from .mixin import OriginMixin


class AddWebAuthnKeyView(OriginMixin, TemplateView):
    template_name = "kagi/add_key.html"

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs["form"] = KeyRegistrationForm()
        return kwargs


class KeyManagementView(ListView):
    template_name = "kagi/key_list.html"

    def get_queryset(self):
        return self.request.user.webauthn_keys.all()

    def post(self, request):
        assert "delete" in self.request.POST
        key = get_object_or_404(self.get_queryset(), pk=self.request.POST["key_id"])
        key.delete()
        messages.success(request, _("Key removed."))
        return HttpResponseRedirect(reverse("kagi:webauthn-keys"))



from io import StringIO
from unittest import mock

from django.contrib.auth.models import User
from django.core.management import CommandError, call_command
from django.urls import reverse

import pytest

from ..models import BackupCode


def test_list_backup_codes(admin_client):
    response = admin_client.get(reverse("kagi:backup-codes"))
    assert list(response.context_data["backupcode_list"]) == []
    assert response.status_code == 200


def test_add_new_backup_codes(admin_client):
    response = admin_client.post(reverse("kagi:backup-codes"))
    assert response.status_code == 302
    response = admin_client.get(reverse("kagi:backup-codes"))
    assert len(response.context_data["backupcode_list"]) == 10


@pytest.mark.django_db
def test_addbackupcode_command():
    User.objects.create(username="admin", password="admin", email="john.doe@kagi.com")
    assert BackupCode.objects.count() == 0
    stdout = StringIO()
    call_command("addbackupcode", "admin", stdout=stdout)
    assert len(stdout.getvalue()) == 8
    assert BackupCode.objects.count() == 1


@pytest.mark.django_db
def test_addbackupcode_command_can_use_a_specific_code():
    User.objects.create(username="admin", password="admin", email="john.doe@kagi.com")
    assert BackupCode.objects.count() == 0
    call_command("addbackupcode", "admin", "--code", "123456", stdout=StringIO())
    assert BackupCode.objects.count() == 1


@pytest.mark.django_db
def test_addbackupcode_command_refuse_to_create_twice_the_same_code():
    User.objects.create(username="admin", password="admin", email="john.doe@kagi.com")
    assert BackupCode.objects.count() == 0
    call_command("addbackupcode", "admin", stdout=StringIO())
    code = BackupCode.objects.get().code
    with pytest.raises(CommandError):
        call_command("addbackupcode", "admin", "--code", code)


@pytest.mark.django_db
def test_backup_code_manager_handles_code_duplication():
    user = User.objects.create(
        username="admin", password="admin", email="john.doe@kagi.com"
    )
    assert BackupCode.objects.count() == 0
    with mock.patch(
        "kagi.models.get_random_string", side_effect=["123456", "123456", "45678"]
    ) as mocked:
        user.backup_codes.create_backup_code()
        user.backup_codes.create_backup_code()

    assert mocked.call_count == 3
    assert BackupCode.objects.count() == 2


@pytest.mark.django_db
def test_a_user_with_backup_codes_and_no_mfa_is_not_asked_for_its_code(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.backup_codes.create_backup_code()
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:two-factor-settings")

    # Are we truly logged in?
    response = client.get(reverse("kagi:two-factor-settings"))
    assert response.status_code == 200


@pytest.mark.django_db
def test_a_user_with_mfa_can_use_a_backup_code(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.totp_devices.create()
    user.backup_codes.create_backup_code(code="123456")
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    response = client.post(
        reverse("kagi:verify-second-factor"), {"type": "backup", "code": "123456"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:two-factor-settings")

    # Check that the used code has been removed.
    assert BackupCode.objects.count() == 0

    # Are we truly logged in?
    response = client.get(reverse("kagi:two-factor-settings"))
    assert response.status_code == 200


@pytest.mark.django_db
def test_a_user_cannot_login_with_a_wrong_backup_code(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.totp_devices.create()
    user.backup_codes.create_backup_code(code="123456")
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    response = client.post(
        reverse("kagi:verify-second-factor"), {"type": "backup", "code": "213456"}
    )
    assert response.status_code == 200
    assert response.context_data["forms"]["backup"].errors == {
        "code": ["That is not a valid backup code."]
    }



from django.conf import settings as django_settings
from django.contrib import auth
from django.contrib.auth.decorators import login_required
from django.http import JsonResponse
from django.shortcuts import resolve_url
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.timezone import now
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_http_methods

import webauthn

from .. import settings, util
from ..forms import KeyRegistrationForm
from ..models import WebAuthnKey

# Registration


@login_required
@require_http_methods(["POST"])
def webauthn_begin_activate(request):
    form = KeyRegistrationForm(request.POST)

    if not form.is_valid():
        return JsonResponse({"errors": form.errors}, status=400)

    username = request.user.get_username()
    display_name = request.user.get_full_name()

    challenge = util.generate_challenge(32)
    ukey = util.generate_ukey()

    request.session["key_name"] = form.cleaned_data["key_name"]
    request.session["challenge"] = challenge
    request.session["register_ukey"] = ukey

    make_credential_options = webauthn.WebAuthnMakeCredentialOptions(
        challenge,
        settings.RELYING_PARTY_NAME,
        settings.RELYING_PARTY_ID,
        ukey,
        username,
        display_name,
        settings.WEBAUTHN_ICON_URL,
    )

    return JsonResponse(make_credential_options.registration_dict)


@login_required
@csrf_exempt
@require_http_methods(["POST"])
def webauthn_verify_credential_info(request):
    challenge = request.session["challenge"]
    ukey = request.session["register_ukey"]

    registration_response = request.POST
    trust_anchor_dir = settings.WEBAUTHN_TRUSTED_CERTIFICATES
    trusted_attestation_cert_required = (
        settings.WEBAUTHN_TRUSTED_ATTESTATION_CERT_REQUIRED
    )
    self_attestation_permitted = settings.WEBAUTHN_SELF_ATTESTATION_PERMITTED
    none_attestation_permitted = settings.WEBAUTHN_NONE_ATTESTATION_PERMITTED

    webauthn_registration_response = webauthn.WebAuthnRegistrationResponse(
        settings.RELYING_PARTY_ID,
        util.get_origin(request),
        registration_response,
        challenge,
        trust_anchor_dir,
        trusted_attestation_cert_required,
        self_attestation_permitted,
        none_attestation_permitted,
        uv_required=False,  # User validation
    )

    try:
        webauthn_credential = webauthn_registration_response.verify()
    except Exception as e:
        return JsonResponse({"fail": f"Registration failed. Error: {e}"}, status=400)

    # W3C spec. Step 17.
    #
    # Check that the credentialId is not yet registered to any other user.
    # If registration is requested for a credential that is already registered
    # to a different user, the Relying Party SHOULD fail this registration
    # ceremony, or it MAY decide to accept the registration, e.g. while deleting
    # the older registration.
    credential_id_exists = WebAuthnKey.objects.filter(
        credential_id=webauthn_credential.credential_id.decode("utf-8")
    ).first()
    if credential_id_exists:
        return JsonResponse({"fail": "Credential ID already exists."}, status=400)

    WebAuthnKey.objects.create(
        user=request.user,
        key_name=request.session.get("key_name", ""),
        ukey=ukey,
        public_key=webauthn_credential.public_key.decode("utf-8"),
        credential_id=webauthn_credential.credential_id.decode("utf-8"),
        sign_count=webauthn_credential.sign_count,
    )

    try:
        del request.session["challenge"]
        del request.session["register_ukey"]
        del request.session["key_name"]
    except KeyError:  # pragma: no cover
        pass

    return JsonResponse({"success": "User successfully registered."})


# Login
@require_http_methods(["POST"])
def webauthn_begin_assertion(request):
    challenge = util.generate_challenge(32)
    request.session["challenge"] = challenge

    user = util.get_user(request)

    username = user.get_username()
    display_name = user.get_full_name()

    keys = WebAuthnKey.objects.filter(user=user)

    webauthn_users = []
    for key in keys:
        webauthn_users.append(
            webauthn.WebAuthnUser(
                key.ukey,
                username,
                display_name,
                settings.WEBAUTHN_ICON_URL,
                key.credential_id,
                key.public_key,
                key.sign_count,
                settings.RELYING_PARTY_ID,
            )
        )

    webauthn_assertion_options = webauthn.WebAuthnAssertionOptions(
        webauthn_users, challenge
    )

    return JsonResponse(webauthn_assertion_options.assertion_dict)


@csrf_exempt
@require_http_methods(["POST"])
def webauthn_verify_assertion(request):
    challenge = request.session.get("challenge")
    assertion_response = request.POST
    credential_id = assertion_response.get("id")

    user = util.get_user(request)

    username = user.get_username()
    display_name = user.get_full_name()

    key = WebAuthnKey.objects.filter(credential_id=credential_id, user=user).first()
    if not key:
        return JsonResponse({"fail": "Key does not exist."}, status=400)

    webauthn_user = webauthn.WebAuthnUser(
        key.ukey,
        username,
        display_name,
        settings.WEBAUTHN_ICON_URL,
        key.credential_id,
        key.public_key,
        key.sign_count,
        settings.RELYING_PARTY_ID,
    )

    webauthn_assertion_response = webauthn.WebAuthnAssertionResponse(
        webauthn_user,
        assertion_response,
        challenge,
        util.get_origin(request),
        uv_required=False,  # User Verification
    )

    try:
        sign_count = webauthn_assertion_response.verify()
    except Exception as e:
        return JsonResponse({"fail": f"Assertion failed. Error: {e}"}, status=400)

    # Update counter.
    key.sign_count = sign_count
    key.last_used_at = now()
    key.save()

    try:
        del request.session["kagi_pre_verify_user_pk"]
        del request.session["kagi_pre_verify_user_backend"]
        del request.session["challenge"]
    except KeyError:  # pragma: no cover
        pass

    auth.login(request, user)

    redirect_to = request.POST.get(
        auth.REDIRECT_FIELD_NAME, request.GET.get(auth.REDIRECT_FIELD_NAME, "")
    )
    if not url_has_allowed_host_and_scheme(
        url=redirect_to, allowed_hosts=[request.get_host()]
    ):
        redirect_to = resolve_url(django_settings.LOGIN_REDIRECT_URL)

    return JsonResponse(
        {
            "success": f"Successfully authenticated as {username}",
            "redirect_to": redirect_to,
        }
    )



import datetime
import hmac
import string

from django.conf import settings
from django.db import IntegrityError, models, transaction
from django.utils import timezone
from django.utils.crypto import get_random_string

from .oath import T, totp


class WebAuthnKey(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="webauthn_keys", on_delete=models.CASCADE
    )
    created_at = models.DateTimeField(auto_now_add=True)
    last_used_at = models.DateTimeField(blank=True, null=True)

    key_name = models.CharField(max_length=64)
    public_key = models.TextField(unique=True)
    ukey = models.TextField(unique=True)
    credential_id = models.TextField(unique=True)
    sign_count = models.IntegerField()

    def __str__(self):
        return f"{self.user} - {self.key_name}"


class BackupCodeManager(models.Manager):
    def create_backup_code(self, code=None):
        if code is not None:
            return self.create(code=code)

        while True:
            try:
                with transaction.atomic():
                    code = get_random_string(length=6, allowed_chars=string.digits)
                    return self.create(code=code)
            except IntegrityError:
                pass


class BackupCode(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="backup_codes", on_delete=models.CASCADE
    )
    code = models.CharField(max_length=8)

    class Meta:
        unique_together = [("user", "code")]

    objects = BackupCodeManager()


class TOTPDevice(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="totp_devices", on_delete=models.CASCADE
    )
    created_at = models.DateTimeField(auto_now_add=True)
    last_used_at = models.DateTimeField(null=True)

    key = models.BinaryField()
    # the T value of the most recently-used token. This prevents using the same
    # token twice.
    last_t = models.PositiveIntegerField(null=True)

    def validate_token(self, token):
        step = datetime.timedelta(seconds=30)
        now = timezone.now()
        # the number of time intervals on either side to check
        slop = 1

        times_to_check = [now + i * step for i in range(-slop, slop + 1)]
        # prevent using the same token twice
        if self.last_t is not None:
            times_to_check = [t for t in times_to_check if T(t) > self.last_t]

        token = str(token)
        for t in times_to_check:
            # BinaryField can be a MemoryView, so make sure to send bytes to hmac.
            if hmac.compare_digest(totp(bytes(self.key), t), token):
                self.last_t = T(t)
                return True
        return False



from django.urls import reverse

import kagi.views


def test_get_admin_login_loads_the_form(client):
    response = client.get(reverse("admin:login"))
    assert isinstance(response.context_data["view"], kagi.views.KagiLoginView)


def test_get_admin_login_redirects_to_admin_index_if_already_logged_in(admin_client):
    response = admin_client.get(reverse("admin:login"))
    assert response.status_code == 302
    assert response.url == reverse("admin:index")



import base64
from datetime import datetime
import re

from django.contrib.auth.models import User
from django.urls import reverse
from django.utils import timezone

import pytest

from ..models import TOTPDevice
from ..oath import totp

base32_regexp = re.compile(
    r"^(?:[A-Z2-7]{8})*(?:[A-Z2-7]{2}={6}|[A-Z2-7]{4}={4}|[A-Z2-7]{5}={3}|[A-Z2-7]{7}=)?$"
)


def add_new_totp_device(client, *, url=None, now=None):
    if url is None:
        url = reverse("kagi:add-totp")

    if now is None:
        now = timezone.now()

    response = client.get(reverse("kagi:add-totp"))
    assert response.status_code == 200

    base32_key = response.context_data["base32_key"]
    key = base64.b32decode(base32_key.encode("utf-8"))
    token = totp(key, now)
    response = client.post(url, {"token": token})
    response.token = token
    return response


def test_list_totp_devices(admin_client):
    response = admin_client.get(reverse("kagi:totp-devices"))
    assert list(response.context_data["totpdevice_list"]) == []
    assert response.status_code == 200


def test_add_a_new_totp_device_shows_a_qrcode(admin_client):
    response = admin_client.get(reverse("kagi:add-totp"))
    assert response.status_code == 200
    assert re.match(
        r"<\?xml version='1\.0' encoding='UTF-8'\?>\n<svg (width|height)=\"49mm\" ",
        response.context_data["qr_svg"],
    )


def test_add_a_new_totp_device_context_data_contains_the_base32_key_and_otpauth_link(
    admin_client,
):
    response = admin_client.get(reverse("kagi:add-totp"))
    assert response.status_code == 200
    assert re.match(base32_regexp, response.context_data["base32_key"])
    assert response.context_data["otpauth"] == (
        f"otpauth://totp/testserver:admin?secret={response.context_data['base32_key']}"
        f"&digits=6&issuer=testserver"
    )


def test_add_a_new_totp_device_validates_the_otpauth_code_but_keeps_secret_in_case_of_mismatch(
    admin_client,
):
    response = admin_client.get(reverse("kagi:add-totp"))
    assert response.status_code == 200
    base32_key = response.context_data["base32_key"]

    # Submit the form once, but with an invalid token. The secret should remain
    # the same.
    response = admin_client.post(reverse("kagi:add-totp"), {"token": "123456"})
    assert response.status_code == 200
    assert base32_key == response.context_data["base32_key"]
    form = response.context_data["form"]
    assert form.errors == {"token": ["That token is invalid."]}

    # Submit the form a second time with the correct token. This should work.
    key = base64.b32decode(base32_key.encode("utf-8"))
    token = totp(key, timezone.now())
    response = admin_client.post(reverse("kagi:add-totp"), {"token": token})
    assert response.status_code == 302
    assert response.url == reverse("kagi:totp-devices")
    # Ensure the secret is removed from the user session after adding a device.
    assert "kagi_totp_secret" not in admin_client.session

    response = admin_client.get(reverse("kagi:totp-devices"))
    assert len(response.context_data["totpdevice_list"]) == 1


def test_add_a_new_totp_device_fails_when_secret_is_missing_in_session(admin_client):
    response = admin_client.get(reverse("kagi:add-totp"))
    assert response.status_code == 200

    session = admin_client.session
    del session["kagi_totp_secret"]
    session.save()

    response = admin_client.post(reverse("kagi:add-totp"), {"token": "123456"})
    assert response.status_code == 302
    assert response.url == reverse("kagi:add-totp")


def test_add_a_new_totp_device_validates_the_otpauth_code_and_create_the_device_if_valid(
    admin_client,
):
    response = add_new_totp_device(admin_client)
    assert response.status_code == 302
    assert response.url == reverse("kagi:totp-devices")
    # Ensure the secret is removed from the user session after adding a device.
    assert "kagi_totp_secret" not in admin_client.session

    response = admin_client.get(reverse("kagi:totp-devices"))
    assert len(response.context_data["totpdevice_list"]) == 1


def test_prevent_reuse_of_totp_device_code(admin_client):
    response = add_new_totp_device(admin_client)
    assert response.status_code == 302
    device = TOTPDevice.objects.get()
    assert not device.validate_token(response.token)


def test_oath_handle_naive_datetime_objects(admin_client):
    response = add_new_totp_device(admin_client, now=datetime.now())
    assert response.status_code == 302


def test_add_a_new_totp_device_redirects_to_the_next_parameter(admin_client):
    url = reverse("kagi:add-totp")
    next_url = reverse("kagi:two-factor-settings")

    response = add_new_totp_device(admin_client, url=f"{url}?next={next_url}")
    assert response.status_code == 302
    assert response.url == next_url


def test_totp_device_deletion_works(admin_client):
    add_new_totp_device(admin_client)
    device = TOTPDevice.objects.get()

    response = admin_client.post(
        reverse("kagi:totp-devices"), {"delete": "checked", "device_id": device.pk}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:totp-devices")
    assert TOTPDevice.objects.count() == 0


@pytest.mark.django_db
def test_a_user_can_confirm_login_with_its_totp_device(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    key = base64.b32decode("7AQ6YRY4OEL7IEHQ6FUWSBRX6W4ZYURF")
    user.totp_devices.create(key=key)
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    now = timezone.now()
    token = totp(key, now)
    response = client.post(
        reverse("kagi:verify-second-factor"), {"type": "totp", "token": token}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:two-factor-settings")

    # Are we truly logged in?
    response = client.get(reverse("kagi:two-factor-settings"))
    assert response.status_code == 200


@pytest.mark.django_db
def test_a_user_cannot_confirm_login_with_a_wrong_token(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    key = base64.b32decode("7AQ6YRY4OEL7IEHQ6FUWSBRX6W4ZYURF")
    user.totp_devices.create(key=key)
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    assert response.url == reverse("kagi:verify-second-factor")

    response = client.post(
        reverse("kagi:verify-second-factor"), {"type": "totp", "token": "123456"}
    )
    assert response.status_code == 200
    assert response.context_data["forms"]["totp"].errors == {
        "token": ["That token is invalid."]
    }


@pytest.mark.django_db
def test_a_user_can_be_redirected_to_the_next_parameter(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    key = base64.b32decode("7AQ6YRY4OEL7IEHQ6FUWSBRX6W4ZYURF")
    user.totp_devices.create(key=key)
    url = reverse("kagi:login")
    next_url = reverse("kagi:add-webauthn-key")

    response = client.post(
        f"{url}?next={next_url}", {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    url = reverse("kagi:verify-second-factor")
    verify_url = f"{url}?next=%2Fkagi%2Fadd-webauthn-key%2F"
    assert response.url == verify_url

    now = timezone.now()
    token = totp(key, now)
    response = client.post(verify_url, {"type": "totp", "token": token})
    assert response.status_code == 302
    assert response.url == reverse("kagi:add-webauthn-key")


@pytest.mark.django_db
def test_an_admin_user_shows_the_admin_base_style(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    user.is_staff = True
    user.is_superuser = True
    user.save()
    key = base64.b32decode("7AQ6YRY4OEL7IEHQ6FUWSBRX6W4ZYURF")
    user.totp_devices.create(key=key)
    url = reverse("admin:login")
    next_url = reverse("kagi:add-webauthn-key")

    response = client.post(
        f"{url}?next={next_url}", {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    url = reverse("kagi:verify-second-factor")
    verify_url = f"{url}?next=%2Fkagi%2Fadd-webauthn-key%2F&admin=1"
    assert response.url == verify_url

    response = client.get(verify_url)
    assert response.status_code == 200

    now = timezone.now()
    token = totp(key, now)
    response = client.post(verify_url, {"type": "totp", "token": token})
    assert response.status_code == 302
    assert response.url == reverse("kagi:add-webauthn-key")


@pytest.mark.django_db
def test_in_case_the_session_expires_redirect_to_login(client):
    # We need to create a couple of WebAuthnKey for our user.
    user = User.objects.create_user("admin", "john.doe@kagi.com", "admin")
    key = base64.b32decode("7AQ6YRY4OEL7IEHQ6FUWSBRX6W4ZYURF")
    user.totp_devices.create(key=key)
    response = client.post(
        reverse("kagi:login"), {"username": "admin", "password": "admin"}
    )
    assert response.status_code == 302
    client.session.flush()
    response = client.get(reverse("kagi:verify-second-factor"))
    assert response.status_code == 302
    assert response.url == reverse("kagi:login")



import os.path

from django.conf import settings

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


RELYING_PARTY_ID = getattr(settings, "RELYING_PARTY_ID", "localhost")
RELYING_PARTY_NAME = getattr(settings, "RELYING_PARTY_NAME", "Kagi Test Project")
WEBAUTHN_ICON_URL = getattr(settings, "WEBAUTHN_ICON_URL", None)
WEBAUTHN_TRUSTED_CERTIFICATES = getattr(
    settings,
    "WEBAUTHN_TRUSTED_CERTIFICATES",
    os.path.join(BASE_DIR, "..", "trusted_attestation_roots"),
)
WEBAUTHN_TRUSTED_ATTESTATION_CERT_REQUIRED = getattr(
    settings, "WEBAUTHN_TRUSTED_ATTESTATION_CERT_REQUIRED", False
)
WEBAUTHN_SELF_ATTESTATION_PERMITTED = getattr(
    settings, "WEBAUTHN_SELF_ATTESTATION_PERMITTED", False
)
WEBAUTHN_NONE_ATTESTATION_PERMITTED = getattr(
    settings, "WEBAUTHN_NONE_ATTESTATION_PERMITTED", False
)



from django.apps import AppConfig


class KagiConfig(AppConfig):
    name = "kagi"
    default_auto_field = "django.db.models.AutoField"

    def monkeypatch_login_view(self):
        from .admin import monkeypatch_admin

        monkeypatch_admin()

    def ready(self):
        self.monkeypatch_login_view()



from django.http import HttpResponseRedirect
from django.views.generic import ListView


class BackupCodesView(ListView):
    template_name = "kagi/backup_codes.html"

    def get_queryset(self):
        return self.request.user.backup_codes.all()

    def post(self, request):
        for i in range(10):
            self.request.user.backup_codes.create_backup_code()
        return HttpResponseRedirect(self.request.build_absolute_uri())



import hashlib
import hmac
import struct


def hotp(key, counter, digits=6):
    """
    These test vectors come from RFC-4226
    (https://tools.ietf.org/html/rfc4226#page-32).

    >>> key = b'12345678901234567890'
    >>> for c in range(10):
    ...     hotp(key, c)
    '755224'
    '287082'
    '359152'
    '969429'
    '338314'
    '254676'
    '287922'
    '162583'
    '399871'
    '520489'
    """

    msg = struct.pack(">Q", counter)
    hs = hmac.new(key, msg, hashlib.sha1).digest()
    offset = hs[19] & 0x0F
    val = struct.unpack(">L", hs[offset : offset + 4])[0] & 0x7FFFFFFF
    return "{val:0{digits}d}".format(val=val % 10**digits, digits=digits)


def T(t, step=30):
    """
    The TOTP T value (number of time steps since the epoch)
    """
    return int(t.timestamp()) // step


def totp(key, t, digits=6, step=30):
    """
    These test vectors come from RFC-6238
    (https://tools.ietf.org/html/rfc6238#appendix-B).
    >>> import datetime
    >>> key = b'12345678901234567890'
    >>> totp(key, datetime.datetime.fromtimestamp(59), digits=8)
    '94287082'
    >>> totp(key, datetime.datetime.fromtimestamp(1111111109), digits=8)
    '07081804'
    >>> totp(key, datetime.datetime.fromtimestamp(20000000000), digits=8)
    '65353130'
    """
    return hotp(key, T(t, step), digits)
